/*	signal.c

	AFAPI signal definitions for TTL_PAPERS.

	July 1996 by H. Dietz
*/


#include "afapi.h"
#include "inline.h"


#define	S_SIGMAX	3


static p_8u s_outlast;

/*	Pre-defined signal handlers...
*/

void
s_ignore(register p_bar m)
{
	/* Nothing to do...  that's why it's called s_ignore() */
}

/*	The RSM handler...

	Really pretty simple stuff.  A pre-installed signal handler is
	defined (s_updateHandler()) as a static function.  Everything
	else is just simple interface crud....
*/

struct s_updateArgsType s_updateArgs;	/* Where the args go... */


/*	The following stuff is used to minimize the quantity of stuff
	that must be transmitted for each update of RSM.  Basically,
	for each object update, we must transmit the object base
	address, size, and value.  The block sends do a nice job of
	optimizing the value transmission, and the sizes usually fit
	within a single nybble, but sending the address is costly....

	The solution here uses a 4-bit tag to identify how the address
	is encoded.  If we are lucky, the tag is the cache entry that
	hit.  Less lucky, it could be the same 256-byte page as a
	cache entry, in which case we just send the 8-bit offset.  If
	that didn't work, we just send the pointer itself.
*/

#define	HTABSIZE	7

static p_32u	haddr[HTABSIZE] = { 0, 0, 0, 0, 0, 0, 0 };

#define	HBITS		8
#define	HMASK		0xff
#define	HPAGE(x)	(((p_32u) x) & ~HMASK)
#define	HHASH(x)	((((p_32u) x) >> HBITS) % HTABSIZE)

inline extern void
s__updatePut()
{
	register void *addr = s_updateArgs.a;
	register p_32 size = s_updateArgs.n;

	if (size == 0) {
		/* My write was squashed by a race */
		p__put(8 | 7);
	} else {
		/* My write is still valid... */
		register p_32u page = HPAGE(addr);
		register p_32u h = HHASH(addr);

		/* Send encoded addr... */
		if (haddr[h] == ((p_32u) addr)) {
			/* Literal match on hashed cache */
			p__put(8 | h);
		} else if (HPAGE(haddr[h]) == page) {
			/* Page match, HBITS-bit page offset... */
			p__put(h);
			p_bcastPut8u(((p_32u) addr) & 0xff);
			haddr[h] = ((p_32u) addr);
		} else {
			/* Page not in table...
			   send 32-bit address and update cache
			*/
			p__put(7);
			p_bcastPutptr(addr);
			haddr[h] = ((p_32u) addr);
		}

		/* Send encoded size... */
		if (size <= 0xf) {
			/* Size has a value 1-15, fits in 4 bits */
			p__put(size);
		} else {
			/* Size doesn't fit, send 32 bits */
			p__put(0);
			p_bcastPut32u(size);
		}

		/* Send data...
		*/
		p_bcastPutz(addr, size);
	}
}


inline extern void
s__updateGet()
{
	register int addrkey = p__get();
	register void *hisaddr;
	register p_32u hissize;

	/* Only have stuff to get if this write wasn't squashed by a
	   race...
	*/
	if (addrkey != (8 | 7)) {
		/* Was this write a literal address cache hit? */
		if (addrkey & 8) {
			/* Yup, we have the entry */
			hisaddr = ((char *) haddr[addrkey & 7]);
		} else {
			/* Nope.  Perhaps an HBITS-bit offset? */
			if (addrkey != 7) {
				/* Yup, get value & update cache */
				hisaddr = ((void *)
					   ((haddr[addrkey] & ~HMASK) |
					    p_bcastGet8u()));
				haddr[addrkey] = ((p_32u) hisaddr);
			} else {
				/* Nope, get full 32 bit address
				   and update cache...
				*/
				hisaddr = p_bcastGetptr();
				haddr[HHASH(hisaddr)] = ((p_32u) hisaddr);
			}
		}

		/* Get encoded size... */
		hissize = p__get();
		if (hissize == 0) {
			hissize = p_bcastGet32u();
		}

		/* Does this make my info redundant? */
		{
			register void *myaddr = s_updateArgs.a;

			if ((hisaddr <= myaddr) &&
			    ((hisaddr + hissize) >=
			     (myaddr + s_updateArgs.n))) {
				/* Yup.  I lost the race.  Zap my size. */
				s_updateArgs.n = 0;
			}
		}

		/* Get the data from him into my memory... */
		p_bcastGetz(hisaddr, hissize);
	}
}


static void
s_updateHandler(register p_bar m)
{
	register int who;
	register p_bar bit = 1;

	/* For each writing PE, do the broadcast store operation */
	for (who=0; who<NPROC; ++who) {
		if (bit & m) {
			if (IPROC == who) {
				s__updatePut();
			} else {
				s__updateGet();
			}
		}

		bit <<= 1;
	}
}

/*	s_vector[] array...

	The following is the magic internal structure used to vector
	signals to the correct handlers.  It is specific to TTL_PAPERS.
	The entry for s_updateHandler must be in the S_SIGMEM position
	in the table.
*/

static void (*s_vector[S_SIGMAX])(register p_bar mask) = {
	/* Normal signals... */
	s_updateHandler,
	s_ignore,
	s_ignore
};

/*	s_install()

	Install a signal handler...
*/

int
s_install(register int s,
register void (*f)(p_bar m))
{
	/* Install the signal handler f()... */
	register int which = s;

	/* If s is S_SIGANY, pick one that is s_ignore()... */
	if (which == S_SIGANY) {
		for (which = 0; which<S_SIGMAX; ++which) {
			if (s_vector[which] == s_ignore) break;
		}
	}

	/* Got a valid signal number? */
	if ((which < 0) || (which >= S_SIGMAX)) {
		/* Nope.  Return with error. */
		return(-1);
	}

	/* Ok, install it... */
	s_vector[which] = f;

	/* Return signal number */
	return(which);
}

/*	s_init() and s_exit()...
*/

int
s_init(void)
{
	/* make sure any previous ops are really done...
	 * since a quirk of using a CAPERS/PAPERSjr makes
	 * it possible for PE0 to miss a data value if PE1
	 * just changes his port...i.e. PE0 must initiate changes
	 */
	/* Flip the strobe bits and signal idle... */
	p__swap(0x00);
	s_outlast = 0x00;

	/* Ok, it worked...? */
	return(0);
}


void
s_exit(void)
{
	/* Tell everybody we are idle and want to exit... */
	register p_8u t = ((p_outlast & 0xf0) | (s_outlast = 0x08));
	register p_8u pin;
	p__outb(t, p_outport);
	p_outlast = t;

	/* Wait for everybody to agree that the signal system
	   should be exited, polling as we wait
	*/
	do {
		pin = p__inb(p_inport);

		if (pin & 0x38) {
			s_signal(S_SIGANY);
		}
	} while (!(pin & 0x40));

	/* To be sure everyone knows, do a real barrier... */
	p__wait();
}

/*	s_signal()...

	This code must recognize who has sent signals, create masks
	grouping similar signals together, and invoke the appropriate
	handler functions in a way that gives RSM the highest priority
	and shared memory exit the lowest priority....

	There are only 4 bits in the TTL_PAPERS data path, and we use
	each bit to define a signal, thus, we have only four signals.
	In fact, one of those signals is reserved for disabling the
	signal system, so only three are available for users.  Signal
	number s is bit (1 << s).

	Later, these 3 signals will probably be expanded by using a
	two-level decoding process for the third signal to yield a
	total of 6 signals, the first 2 of which are "fast."  The
	other choice would be to make 9 user signals with only the
	first (S_SIGMEM) being "fast."  Anyway, for now there are 3
	"fast" user signals....
*/

void
s_signal(register int s)
{
	register unsigned char t, t2;
	register unsigned char match;
	register int result;

	/* A valid signal...? If so, update s_oulast... */
	s_outlast = (((s >= 0) && (s <= S_SIGMAX)) ?
		     (s_outlast |= (1 << s)) :
		     (s_outlast & 0x08));

	/* Paste-in the signal and output it */
	p_outlast = ((p_outlast & 0xf0) | s_outlast);
	p__outb(p_outlast, p_outport);

	/* Now do a barrier synchronization */
	p__wait();

	/* Read the data value again (time for data to settle) and
	   paste-in our signal value...
	*/
	result = (p__inb(p_inport) | (p_outlast << 3));

	/* Another barrier synchronization to make sure we agree */
	p__wait();

	/* Result is now a mask saying which signals...  process them
	   in priority order, shared memory first.  Notice that higher
	   bit positions are lower priority, with disabling the signal
	   system lowest.
	*/
	if (result & 0x08) {
		register p_bar mask = p_waitbar(s == 0);
		s_vector[0](mask);
	}
	if (result & 0x10) {
		register p_bar mask = p_waitbar(s == 1);
		s_vector[1](mask);
	}
	if (result & 0x20) {
		register p_bar mask = p_waitbar(s == 2);
		s_vector[2](mask);
	}

	/* Yet another barrier synchronization */
	/* Now reset any signals we have processed... */
	p__swap(s_outlast &= 0x08);
}

/*	Inlined signal and RSM stuff...

	These are just callable versions of the simple interface
	routines....
*/

void
s_poll(void)
{
	s__poll();
}

void
s_wait(void)
{
	s__wait();
}


void
s_update(register void *a,
register int n)
{
	s__update(a, n);
}


void
s_initLock(s_lock *lock)
{
	s__initLock(lock);
}


void
s_acquire(s_lock *lock)
{
	s__acquire(lock);
}


void
s_release(s_lock *lock)
{
	s__release(lock);
}
